# Author: Courtney Lynn Zambory (courtney.zambory@gmail.com)
# Last date of Update: May 14, 2018
# Works with python version 2.7.10
# This script uses LiDAR Digital Elevation Models (DEMs) to find riparian depressions on the landcape
# and characterize their topographic and morphological characteristics

# Import Modules
import arcpy
from arcpy import env
from arcpy.sa import *
import sys, string, os, time
from subprocess import call
import multiprocessing
arcpy.CheckOutExtension("Spatial")
import math


from arcpy.sa import *
arcpy.CheckOutExtension('Spatial')

# Set Paths and Environments
arcpy.env.overwriteOutput = True


# Define Parameters
DEM = arcpy.GetParameterAsText(0)
meters = arcpy.GetParameterAsText(1)
PStream = arcpy.GetParameterAsText(2)
all_streams = arcpy.GetParameterAsText(3)
size_limit = arcpy.GetParameterAsText(4)
rele = arcpy.GetParameterAsText(5)
ZFactor = arcpy.GetParameterAsText(6)
output = arcpy.GetParameterAsText(7)
riparian_zone = arcpy.Buffer_analysis(PStream, 'riparian_zone', meters, 'FULL', 'ROUND', 'NONE', '')
Fill_DEM = Fill(DEM)

# Extract the DEMs to the 500yr Flood Zone
rDEM = ExtractByMask(DEM, riparian_zone)
rFilled_DEM = ExtractByMask(Fill_DEM, riparian_zone)


arcpy.AddMessage('Starting process...')
arcpy.AddMessage("Limiting search for depressions to %s meters around perennial streams." % (meters))


arcpy.AddMessage('Subtracting DEM from Filled DEM to find Depressions...')
OrgiDepressions = Minus(rFilled_DEM, rDEM)
arcpy.CopyRaster_management(OrgiDepressions, 'in_memory' + '//orgidepress,tif')
arcpy.AddMessage ('Converting all areas without sinks to value of no data while sink areas are values of 1')
Convert = Con(OrgiDepressions, 1, '', 'value > 0')
arcpy.AddMessage ('Converting sink areas (areas with a value of 1) to polygon shapefiles...')
all_depress = arcpy.RasterToPolygon_conversion(Convert, 'all_depress')
Copy = arcpy.CopyFeatures_management(all_depress, 'in_memory' + '//all_depress')
arcpy.Delete_management(Convert)
arcpy.AddMessage ('Calculating the area of all new polygons...')
arcpy.AddField_management(all_depress, "ShapeAreaf", 'DOUBLE')
arcpy.CalculateField_management(all_depress, "ShapeAreaf", "!shape.area@squaremeters!", 'PYTHON')
arcpy.AddField_management(all_depress, "depress_ID", 'INTEGER')
Copy = arcpy.CopyFeatures_management(all_depress, 'in_memory' + '//copy_feature')
arcpy.CalculateField_management(Copy, "depress_ID", "!OBJECTID!", 'PYTHON')
first_count = arcpy.GetCount_management(Copy)
arcpy.AddMessage('Initially, ' + str(first_count) + ' depressions are found within the riparian zone.')          
# Delete any depressions outside of size contraints
arcpy.AddMessage ('Deleting features that are less than %s square meters...' % (size_limit))
with arcpy.da.UpdateCursor(Copy, "ShapeAreaf") as rows:
    for each_row in rows:
        if each_row[0] < float(size_limit):
            rows.deleteRow()
two_count = arcpy.GetCount_management(Copy)
arcpy.AddMessage('After, ' + str(two_count) + ' depressions are found within the riparian zone.') 
Copy2 = arcpy.CopyFeatures_management(Copy, 'in_memory' + '//copy_feature2')
# Delete any Depressions that intersect the Stream Channel
arcpy.AddMessage ('Deleting features that intersect the stream channel...')
no_str_depress = arcpy.MakeFeatureLayer_management(Copy, "no_str_depress")
arcpy.SelectLayerByLocation_management(no_str_depress, 'INTERSECT', all_streams, '', 'NEW_SELECTION', 'INVERT' )
no_stream_depress = arcpy.CopyFeatures_management(no_str_depress, 'in_memory' + '//no_stream_depress')

    ###########################################################
    #Generate Minimum Convex Bounding Geometry
    ###########################################################

arcpy.AddMessage('Generating Bounding Geometries (convex bounding)...')
boundaries_a = arcpy.MinimumBoundingGeometry_management(no_stream_depress, 'boundaries_a', 'CONVEX_HULL', 'NONE')
arcpy.AddField_management(boundaries_a, "convex_a", 'DOUBLE')
arcpy.AddField_management(boundaries_a, "convex_p", 'DOUBLE')
arcpy.CalculateField_management(boundaries_a, "convex_a", "!shape.area@squaremeters!", 'PYTHON')
arcpy.CalculateField_management(boundaries_a, "convex_p", "!shape.geodesicLength@METERS!", 'PYTHON')
FieldListName = ['Area']
arcpy.DeleteField_management(boundaries_a, FieldListName)
#Create a table of the convex bounding shapefile
arcpy.AddMessage('Creating table of convex bounding shapefile...')
temp_table = arcpy.MakeTableView_management(boundaries_a, 'temp_table')
final_table = arcpy.CopyRows_management(temp_table, 'final_table')
#Spatial Join Minimum Convex Bounding Geometry and Depression Features
arcpy.AddMessage('Joining layers...') 
Oxbows_test = arcpy.JoinField_management(no_stream_depress, "depress_ID", final_table, "depress_ID", '')
Oxbows_orig = arcpy.CopyFeatures_management(Oxbows_test, 'Oxbows_orig')
#Delete Unnesssary Fields
FieldListName = ['ORIG_FID', 'OID', 'ID', 'ID_1', 'GRIDCODE_1', 'depress__1']
arcpy.DeleteField_management(Oxbows_orig, FieldListName)
Oxbows_a = arcpy.CopyFeatures_management(Oxbows_orig, 'Oxbows_a')
arcpy.AddField_management(Oxbows_a, "perim", 'DOUBLE')
arcpy.CalculateField_management(Oxbows_a, "perim", "!shape.geodesicLength@METERS!", 'PYTHON')

    ###########################################################
    #Generate Minimum RECTANGLE_BY_WIDTH Bounding Geometry
    ###########################################################

boundaries_b = arcpy.MinimumBoundingGeometry_management(Oxbows_a, 'boundaries_b', 'RECTANGLE_BY_WIDTH', '', '', 'MBG_FIELDS')
FieldListName = ['ORIG_FID', 'ID_1', 'GRIDCODE', 'depress__1', 'GRIDCODE_1', 'convex_a', 'convex_p', 'perimeter']
arcpy.DeleteField_management(boundaries_b, FieldListName)
temp_table = arcpy.MakeTableView_management(boundaries_b, 'temp_table')
final_table = arcpy.CopyRows_management(temp_table, 'final_table')
#Spatial Join Minimum Convex Bounding Geometry and Depression Features
arcpy.AddMessage('Joining layers...') 
Oxbows_test = arcpy.JoinField_management(Oxbows_a, "depress_ID", final_table, "depress_ID", '')
Oxbows_shape = arcpy.CopyFeatures_management(Oxbows_test, 'Oxbows_shape')
arcpy.AddField_management(Oxbows_shape, "CIR", 'DOUBLE')
arcpy.AddField_management(Oxbows_shape, "SLD", 'DOUBLE')
arcpy.AddField_management(Oxbows_shape, "REC", 'DOUBLE')
arcpy.AddField_management(Oxbows_shape, "CVX", 'DOUBLE')
arcpy.AddField_management(Oxbows_shape, "CON", 'DOUBLE')
arcpy.CalculateField_management(Oxbows_shape, "CIR", "!perim!/(2*(math.sqrt(3.14 * !ShapeAreaf!)))", 'PYTHON')
arcpy.CalculateField_management(Oxbows_shape, "SLD", "!ShapeAreaf! / !convex_a!", 'PYTHON')
arcpy.CalculateField_management(Oxbows_shape, "REC", "!ShapeAreaf! / ((2* !MBG_Length!)*(2* !MBG_Width!))", 'PYTHON')
arcpy.CalculateField_management(Oxbows_shape, "CVX", "!convex_p!/!perim!", 'PYTHON')
arcpy.CalculateField_management(Oxbows_shape, "CON", "math.sqrt((!SLD! * !SLD!) + (!CVX! * !CVX!))", 'PYTHON')
Oxbows = Oxbows_shape



rslope = Slope(rDEM, "DEGREE", ZFactor)
neighborhood = NbrRectangle(5, 5, "CELL")
focal_stats = FocalStatistics(rslope, neighborhood, "STD","")
arcpy.AddMessage('Calculating mean depth of the depressions in cm....')
Stats = ZonalStatisticsAsTable(Oxbows, "depress_ID", OrgiDepressions, "in_memory" + "\\Stats1", "", "MEAN")
arcpy.JoinField_management(Oxbows, "depress_ID", Stats, "depress_ID", ["MEAN"])
arcpy.AddField_management(Oxbows, "MEANDepth", 'DOUBLE')
arcpy.CalculateField_management(Oxbows, "MEANDepth", "(!MEAN! * (%s * 100))" % ZFactor, "PYTHON")
arcpy.DeleteField_management(Oxbows, ["gridcode", "MEAN", 'Shape_Ar_1'])

arcpy.AddMessage('Calculating maximum depth of the depressions in cm....')

Stats = ZonalStatisticsAsTable(Oxbows, "depress_ID", OrgiDepressions, "in_memory" + "\\Stats2", "", "MAXIMUM")
arcpy.JoinField_management(Oxbows, "depress_ID", Stats, "depress_ID", ["MAX"])
arcpy.AddField_management(Oxbows, "MAXDepth", 'DOUBLE')
arcpy.CalculateField_management(Oxbows, "MAXDepth", "(!MAX! * (%s * 100))" % ZFactor, "PYTHON")

arcpy.AddMessage('Calculating mean HAC of the depressions in cm....')

Connectivity = ZonalStatisticsAsTable(Oxbows, "depress_ID", rele, "in_memory" + "\\connect1", "", "MEAN")
arcpy.JoinField_management(Oxbows, "depress_ID", Connectivity, "depress_ID", ["MEAN"])
arcpy.AddField_management(Oxbows, "MEANHAC", 'DOUBLE')
arcpy.CalculateField_management(Oxbows, "MEANHAC", "(!MEAN! * (%s * 100))" % ZFactor, "PYTHON")
arcpy.DeleteField_management(Oxbows, ["gridcode", "MEAN", 'Shape_Ar_1'])

arcpy.AddMessage('Calculating mean slope of the depressions in degrees....')

Connectivity = ZonalStatisticsAsTable(Oxbows, "depress_ID", rslope, "in_memory" +"\\connect2", "", "MEAN")
arcpy.JoinField_management(Oxbows, "depress_ID", Connectivity, "depress_ID", ["MEAN"])
arcpy.AddField_management(Oxbows, "SLOPE", 'DOUBLE')
arcpy.CalculateField_management(Oxbows, "SLOPE", "(!MEAN!)", "PYTHON")
arcpy.DeleteField_management(Oxbows, ["gridcode", "MEAN", 'Shape_Ar_1'])

arcpy.AddMessage('Calculating mean roughness of the depressions....')

Connectivity = ZonalStatisticsAsTable(Oxbows, "depress_ID", focal_stats, "in_memory" +"\\connect3", "", "MEAN")
arcpy.JoinField_management(Oxbows, "depress_ID", Connectivity, "depress_ID", ["MEAN"])
arcpy.AddField_management(Oxbows, "Roughness", 'DOUBLE')
arcpy.CalculateField_management(Oxbows, "Roughness", "(!MEAN!)", "PYTHON")
FieldListName = ['depress__1', 'convex_a', 'convex_p', 'depress__2', 'depress__3', 'MBG_Length','MBG_Width', 'MBG_Orient', 'MEAN', 'MEAN_1', 'MEAN_12','MBG_Orient', 'perim_1', 'Shape_Le_1', 'MAX', 'Shape_Le_2', 'Shape_Le_3', 'Shape_Le_4', 'Shape_Le_5', 'Shape_Le_6', 'Shape_Le_7', 'ShapeAre_1', 'ShapeAre_2', 'ShapeAre_3']
arcpy.DeleteField_management(Oxbows, FieldListName)

fields = ['SLD', 'MEANHAC', 'CON', 'Roughness', 'CIR', 'MEANDepth', 'SLOPE', 'MAXDepth', 'perim', 'CVX', 'Rank']
#fields = ['0',      '1',      '2',     '3',      '4',     '5',        '6',       '7',      '8',    '9',   '10']

arcpy.AddField_management(Oxbows, "Rank", 'DOUBLE')
with arcpy.da.UpdateCursor(Oxbows, fields) as cursor:
    for row in cursor:
        if (row[0] >= 0.5998 and row[0] < 0.7149 and row[5] >= 9.034 and row[4] >= 1.739 and row[2] < 0.4256 and row[8] >= 613.7):
            row[10] = 2
        elif (row[0] >= 0.5998 and row[0] < 0.7149 and row[5] >= 9.034 and row[4] >= 1.739 and row[2] >= 0.4256):
            row[10] = 0
        elif (row[0] >= 0.5998 and row[0] < 0.7149 and row[5] >= 9.034 and row[4] < 1.739):
            row[10] = 1
        elif (row[0] < 0.5998 and row[1] >= 223.2 and row[0] < 0.4092 and row[1] < 424.8):
            row[10] = 5
        elif (row[0] < 0.5998 and row[1] < 223.2 and row[2] < 0.4805 and row[3] < 1.385 and row[5] >= 4.651 and row[7] >= 17.5 and row[5] >= 16.13):
            row[10] = 4
        elif (row[0] < 0.5998 and row[1] < 223.2 and row[2] < 0.4805 and row[3] < 1.385 and row[5] >= 4.651 and row[7] < 17.5):
            row[10] = 6
        elif (row[0] < 0.5998 and row[1] < 223.2 and row[2] < 0.4805 and row[3] >= 1.385 and row[9] < 4.651):
            row[10] = 3
        elif (row[0] < 0.5998 and row[1] < 223.2 and row[2] >= 0.4805 and row[6] < 8.182):
            row[10] = 7
        else:
            row[10] = -9
        cursor.updateRow(row)

with arcpy.da.UpdateCursor(Oxbows, "Rank") as rows:
    for each_row in rows:
        if each_row[0] < 0:
            rows.deleteRow()
final_final_oxbows = arcpy.CopyFeatures_management(Oxbows, output)


